@@ -1,9 +1,9 @@
-from rope.base.fscommands import _decode_data
+from typing import Optional
+
 from rope.base import ast, exceptions, utils
 
 
-class PyObject(object):
-
+class PyObject:
     def __init__(self, type_):
         if type_ is None:
             type_ = self
@@ -16,10 +16,12 @@ class PyObject(object):
 
     def get_attribute(self, name):
         if name not in self.get_attributes():
-            raise exceptions.AttributeNotFoundError(
-                'Attribute %s not found' % name)
+            raise exceptions.AttributeNotFoundError("Attribute %s not found" % name)
         return self.get_attributes()[name]
 
+    def get_module(self):
+        return None
+
     def get_type(self):
         return self.type
 
@@ -58,7 +60,7 @@ class PyObject(object):
         if type(self) == PyObject and self != self.type:
             return hash(self.type) + 1
         else:
-            return super(PyObject, self).__hash__()
+            return super().__hash__()
 
     def __iter__(self):
         """The same as ``iter(self.get_attributes())``"""
@@ -72,10 +74,10 @@ class PyObject(object):
         if PyObject._types is None:
             PyObject._types = {}
             base_type = PyObject(None)
-            PyObject._types['Type'] = base_type
-            PyObject._types['Module'] = PyObject(base_type)
-            PyObject._types['Function'] = PyObject(base_type)
-            PyObject._types['Unknown'] = PyObject(base_type)
+            PyObject._types["Type"] = base_type
+            PyObject._types["Module"] = PyObject(base_type)
+            PyObject._types["Function"] = PyObject(base_type)
+            PyObject._types["Unknown"] = PyObject(base_type)
         return PyObject._types[name]
 
 
@@ -114,14 +116,13 @@ def get_unknown():
 
     """
     if PyObject._unknown is None:
-        PyObject._unknown = PyObject(get_base_type('Unknown'))
+        PyObject._unknown = PyObject(get_base_type("Unknown"))
     return PyObject._unknown
 
 
 class AbstractClass(PyObject):
-
     def __init__(self):
-        super(AbstractClass, self).__init__(get_base_type('Type'))
+        super().__init__(get_base_type("Type"))
 
     def get_name(self):
         pass
@@ -134,9 +135,8 @@ class AbstractClass(PyObject):
 
 
 class AbstractFunction(PyObject):
-
     def __init__(self):
-        super(AbstractFunction, self).__init__(get_base_type('Function'))
+        super().__init__(get_base_type("Function"))
 
     def get_name(self):
         pass
@@ -152,9 +152,8 @@ class AbstractFunction(PyObject):
 
 
 class AbstractModule(PyObject):
-
     def __init__(self, doc=None):
-        super(AbstractModule, self).__init__(get_base_type('Module'))
+        super().__init__(get_base_type("Module"))
 
     def get_doc(self):
         pass
@@ -163,7 +162,7 @@ class AbstractModule(PyObject):
         pass
 
 
-class PyDefinedObject(object):
+class PyDefinedObject:
     """Python defined names that rope can access their sources"""
 
     def __init__(self, pycore, ast_node, parent):
@@ -176,6 +175,19 @@ class PyDefinedObject(object):
         self.attributes = self.get_module()._get_concluded_data()
         self.defineds = None
 
+    def __repr__(self):
+        return '<{}.{} "{}" at {}>'.format(
+            self.__class__.__module__,
+            self.__class__.__name__,
+            self.absolute_name,
+            hex(id(self)),
+        )
+
+    @property
+    def absolute_name(self):
+        obj_name = self.get_name()
+        return self.get_module().get_name() + ("::" + obj_name if obj_name else "")
+
     visitor_class = None
 
     @utils.prevent_recursion(lambda: {})
@@ -203,8 +215,7 @@ class PyDefinedObject(object):
             return self._get_structural_attributes()[name]
         if name in self._get_concluded_attributes():
             return self._get_concluded_attributes()[name]
-        raise exceptions.AttributeNotFoundError('Attribute %s not found' %
-                                                name)
+        raise exceptions.AttributeNotFoundError("Attribute %s not found" % name)
 
     def get_scope(self):
         if self.scope is None:
@@ -217,14 +228,19 @@ class PyDefinedObject(object):
             current_object = current_object.parent
         return current_object
 
-    def get_doc(self):
+    def get_name(self):
+        return None
+
+    def get_doc(self) -> Optional[str]:
         if len(self.get_ast().body) > 0:
             expr = self.get_ast().body[0]
-            if isinstance(expr, ast.Expr) and \
-               isinstance(expr.value, ast.Str):
-                docstring = expr.value.s
-                coding = self.get_module().coding
-                return _decode_data(docstring, coding)
+            if (
+                isinstance(expr, ast.Expr)
+                and isinstance(expr.value, ast.Constant)
+                and isinstance(expr.value.value, str)
+            ):
+                return expr.value.value
+        return None
 
     def _get_defined_objects(self):
         if self.defineds is None:
@@ -235,8 +251,8 @@ class PyDefinedObject(object):
         if self.visitor_class is None:
             return {}
         new_visitor = self.visitor_class(self.pycore, self)
-        for child in ast.get_child_nodes(self.ast_node):
-            ast.walk(child, new_visitor)
+        for child in ast.iter_child_nodes(self.ast_node):
+            new_visitor.visit(child)
         self.defineds = new_visitor.defineds
         return new_visitor.names
 
@@ -251,15 +267,21 @@ class PyDefinedObject(object):
 
 
 class PyFunction(PyDefinedObject, AbstractFunction):
-    """Only a placeholder"""
+    pass
+
+
+class PyComprehension(PyDefinedObject, PyObject):
+    pass
+
+    def get_name(self):
+        return "<comprehension>"
 
 
 class PyClass(PyDefinedObject, AbstractClass):
-    """Only a placeholder"""
+    pass
 
 
-class _ConcludedData(object):
-
+class _ConcludedData:
     def __init__(self):
         self.data_ = None
 
@@ -275,17 +297,20 @@ class _ConcludedData(object):
         self.data = None
 
     def __str__(self):
-        return '<' + str(self.data) + '>'
+        return "<" + str(self.data) + ">"
 
 
 class _PyModule(PyDefinedObject, AbstractModule):
-
     def __init__(self, pycore, ast_node, resource):
         self.resource = resource
         self.concluded_data = []
         AbstractModule.__init__(self)
         PyDefinedObject.__init__(self, pycore, ast_node, None)
 
+    @property
+    def absolute_name(self) -> str:
+        return self.get_name()
+
     def _get_concluded_data(self):
         new_data = _ConcludedData()
         self.concluded_data.append(new_data)
@@ -300,11 +325,11 @@ class _PyModule(PyDefinedObject, AbstractModule):
 
 
 class PyModule(_PyModule):
-    """Only a placeholder"""
+    pass
 
 
 class PyPackage(_PyModule):
-    """Only a placeholder"""
+    pass
 
 
 class IsBeingInferredError(exceptions.RopeError):
