@@ -1,13 +1,13 @@
 import re
 from contextlib import contextmanager
 from itertools import chain
+from typing import Dict
 
 from rope.base import ast, codeanalyze
-from rope.base.change import ChangeSet, ChangeContents
+from rope.base.change import ChangeContents, ChangeSet
 from rope.base.exceptions import RefactoringError
-from rope.base.utils import pycompat
 from rope.base.utils.datastructures import OrderedSet
-from rope.refactor import sourceutils, similarfinder, patchedast, suites, usefunction
+from rope.refactor import patchedast, similarfinder, sourceutils, suites, usefunction
 
 
 # Extract refactoring has lots of special cases.  I tried to split it
@@ -34,9 +34,8 @@ from rope.refactor import sourceutils, similarfinder, patchedast, suites, usefun
 #
 # There are a few more helper functions and classes used by above
 # classes.
-class _ExtractRefactoring(object):
-
-    kind_prefixes = {}
+class _ExtractRefactoring:
+    kind_prefixes: Dict[str, str] = {}
 
     def __init__(self, project, resource, start_offset, end_offset, variable=False):
         self.project = project
@@ -81,7 +80,7 @@ class _ExtractRefactoring(object):
         )
         info.kind = self._get_kind(kind)
         new_contents = _ExtractPerformer(info).extract()
-        changes = ChangeSet("Extract %s <%s>" % (info.kind, extracted_name))
+        changes = ChangeSet(f"Extract {info.kind} <{extracted_name}>")
         changes.add_change(ChangeContents(self.resource, new_contents))
         return changes
 
@@ -99,7 +98,7 @@ class _ExtractRefactoring(object):
 
     @classmethod
     def _get_kind(cls, kind):
-        raise NotImplementedError("You have to sublass {}".format(cls))
+        raise NotImplementedError(f"You have to subclass {cls}")
 
 
 class ExtractMethod(_ExtractRefactoring):
@@ -116,7 +115,7 @@ class ExtractVariable(_ExtractRefactoring):
     def __init__(self, *args, **kwds):
         kwds = dict(kwds)
         kwds["variable"] = True
-        super(ExtractVariable, self).__init__(*args, **kwds)
+        super().__init__(*args, **kwds)
 
     kind = "variable"
 
@@ -124,7 +123,7 @@ class ExtractVariable(_ExtractRefactoring):
         return cls.kind
 
 
-class _ExtractInfo(object):
+class _ExtractInfo:
     """Holds information about the extract to be performed"""
 
     def __init__(
@@ -223,13 +222,13 @@ class _ExtractInfo(object):
     def extracted(self):
         return self.source[self.region[0] : self.region[1]]
 
-    _cached_parsed_extraced = None
+    _cached_parsed_extracted = None
 
     @property
     def _parsed_extracted(self):
-        if self._cached_parsed_extraced is None:
-            self._cached_parsed_extraced = _parse_text(self.extracted)
-        return self._cached_parsed_extraced
+        if self._cached_parsed_extracted is None:
+            self._cached_parsed_extracted = _parse_text(self.extracted)
+        return self._cached_parsed_extracted
 
     _returned = None
 
@@ -251,8 +250,22 @@ class _ExtractInfo(object):
             )
         return self._returning_named_expr
 
+    _returning_generator = None
 
-class _ExtractCollector(object):
+    @property
+    def returning_generator_exp(self):
+        """Does the extracted piece contains a generator expression"""
+        if self._returning_generator is None:
+            self._returning_generator = (
+                isinstance(self._parsed_extracted, ast.Module)
+                and isinstance(self._parsed_extracted.body[0], ast.Expr)
+                and isinstance(self._parsed_extracted.body[0].value, ast.GeneratorExp)
+            )
+
+        return self._returning_generator
+
+
+class _ExtractCollector:
     """Collects information needed for performing the extract"""
 
     def __init__(self, info):
@@ -265,7 +278,7 @@ class _ExtractCollector(object):
         self.definition_location = None
 
 
-class _ExtractPerformer(object):
+class _ExtractPerformer:
     def __init__(self, info):
         self.info = info
         _ExceptionalConditionChecker()(self.info)
@@ -376,7 +389,7 @@ class _ExtractPerformer(object):
         collector.checks = parts.get_checks()
 
 
-class _DefinitionLocationFinder(object):
+class _DefinitionLocationFinder:
     def __init__(self, info, matched_lines):
         self.info = info
         self.matched_lines = matched_lines
@@ -420,7 +433,7 @@ class _DefinitionLocationFinder(object):
         return self.info.scope.get_end() + 1
 
 
-class _ExceptionalConditionChecker(object):
+class _ExceptionalConditionChecker:
     def __call__(self, info):
         self.base_conditions(info)
         if info.one_line:
@@ -431,8 +444,10 @@ class _ExceptionalConditionChecker(object):
     def base_conditions(self, info):
         if info.region[1] > info.scope_region[1]:
             raise RefactoringError("Bad region selected for extract method")
+
         end_line = info.region_lines[1]
         end_scope = info.global_scope.get_inner_scope_for_line(end_line)
+
         if end_scope != info.scope and end_scope.get_end() != end_line:
             raise RefactoringError("Bad region selected for extract method")
         try:
@@ -484,6 +499,14 @@ class _ExceptionalConditionChecker(object):
             raise RefactoringError(
                 "Extracted piece should contain complete statements."
             )
+        unbalanced_region_finder = _UnbalancedRegionFinder(
+            info.region_lines[0], info.region_lines[1]
+        )
+        unbalanced_region_finder.visit(info.pymodule.ast_node)
+        if unbalanced_region_finder.error:
+            raise RefactoringError(
+                "Extracted piece cannot contain the start of a block without the end."
+            )
 
     def _is_region_on_a_word(self, info):
         if (
@@ -501,7 +524,7 @@ class _ExceptionalConditionChecker(object):
         return next.isalnum() or next == "_"
 
 
-class _ExtractMethodParts(object):
+class _ExtractMethodParts(ast.RopeNodeVisitor):
     def __init__(self, info):
         self.info = info
         self.info_collector = self._create_info_collector()
@@ -576,7 +599,7 @@ class _ExtractMethodParts(object):
         )
         body = self.info.source[self.info.scope_region[0] : self.info.scope_region[1]]
         node = _parse_text(body)
-        ast.walk(node, info_collector)
+        info_collector.visit(node)
         return info_collector
 
     def _get_function_definition(self):
@@ -716,6 +739,7 @@ class _ExtractMethodParts(object):
     def _get_multiline_function_body(self, returns):
         unindented_body = sourceutils.fix_indentation(self.info.extracted, 0)
         unindented_body = self._insert_globals(unindented_body)
+        unindented_body = self._insert_nonlocals(unindented_body)
         if returns:
             unindented_body += "\nreturn %s" % self._get_comma_form(returns)
         return unindented_body
@@ -723,7 +747,9 @@ class _ExtractMethodParts(object):
     def _get_single_expression_function_body(self):
         extracted = _get_single_expression_body(self.info.extracted, info=self.info)
         body = "return " + extracted
-        return self._insert_globals(body)
+        body = self._insert_globals(body)
+        body = self._insert_nonlocals(body)
+        return body
 
     def _insert_globals(self, unindented_body):
         globals_in_body = self._get_globals_in_body(unindented_body)
@@ -738,22 +764,41 @@ class _ExtractMethodParts(object):
             )
         return unindented_body
 
+    def _insert_nonlocals(self, unindented_body):
+        nonlocals_in_body = self._get_nonlocals_in_body(unindented_body)
+        nonlocals_ = self.info_collector.nonlocals_ & (
+            self.info_collector.written | self.info_collector.maybe_written
+        )
+        nonlocals_ = nonlocals_ - nonlocals_in_body
+
+        if nonlocals_:
+            unindented_body = "nonlocal {}\n{}".format(
+                ", ".join(nonlocals_), unindented_body
+            )
+        return unindented_body
+
     @staticmethod
     def _get_globals_in_body(unindented_body):
         node = _parse_text(unindented_body)
         visitor = _GlobalFinder()
-        ast.walk(node, visitor)
+        visitor.visit(node)
         return visitor.globals_
 
+    @staticmethod
+    def _get_nonlocals_in_body(unindented_body):
+        node = _parse_text(unindented_body)
+        visitor = _NonlocalFinder()
+        visitor.visit(node)
+        return visitor.nonlocals_
 
-class _ExtractVariableParts(object):
+
+class _ExtractVariableParts:
     def __init__(self, info):
         self.info = info
 
     def get_definition(self):
         extracted = _get_single_expression_body(self.info.extracted, info=self.info)
-        result = self.info.new_name + " = " + extracted + "\n"
-        return result
+        return self.info.new_name + " = " + extracted + "\n"
 
     def get_body_pattern(self):
         return "(%s)" % self.info.extracted.strip()
@@ -765,7 +810,7 @@ class _ExtractVariableParts(object):
         return {}
 
 
-class _FunctionInformationCollector(object):
+class _FunctionInformationCollector(ast.RopeNodeVisitor):
     def __init__(self, start, end, is_global):
         self.start = start
         self.end = end
@@ -779,6 +824,7 @@ class _FunctionInformationCollector(object):
         self.host_function = True
         self.conditional = False
         self.globals_ = OrderedSet()
+        self.nonlocals_ = OrderedSet()
         self.surrounded_by_loop = 0
         self.loop_depth = 0
 
@@ -810,18 +856,21 @@ class _FunctionInformationCollector(object):
             for name in _get_argnames(node.args):
                 self._written_variable(name, node.lineno)
             for child in node.body:
-                ast.walk(child, self)
+                self.visit(child)
         else:
             self._written_variable(node.name, node.lineno)
             visitor = _VariableReadsAndWritesFinder()
             for child in node.body:
-                ast.walk(child, visitor)
+                visitor.visit(child)
             for name in visitor.read - visitor.written:
                 self._read_variable(name, node.lineno)
 
     def _Global(self, node):
         self.globals_.add(*node.names)
 
+    def _Nonlocal(self, node):
+        self.nonlocals_.add(*node.names)
+
     def _AsyncFunctionDef(self, node):
         self._FunctionDef(node)
 
@@ -834,21 +883,21 @@ class _FunctionInformationCollector(object):
     def _MatchAs(self, node):
         self._written_variable(node.name, node.lineno)
         if node.pattern:
-            ast.walk(node.pattern, self)
+            self.visit(node.pattern)
 
     def _Assign(self, node):
-        ast.walk(node.value, self)
+        self.visit(node.value)
         for child in node.targets:
-            ast.walk(child, self)
+            self.visit(child)
 
     def _AugAssign(self, node):
-        ast.walk(node.value, self)
-        if isinstance(node.target, ast.Subscript):
-            target_id = node.target.value.id
-        else:
+        self.visit(node.value)
+        if isinstance(node.target, ast.Name):
             target_id = node.target.id
-        self._read_variable(target_id, node.target.lineno)
-        self._written_variable(target_id, node.target.lineno)
+            self._read_variable(target_id, node.target.lineno)
+            self._written_variable(target_id, node.target.lineno)
+        else:
+            self.visit(node.target)
 
     def _ClassDef(self, node):
         self._written_variable(node.name, node.lineno)
@@ -870,8 +919,8 @@ class _FunctionInformationCollector(object):
         written = OrderedSet(self.written)
         maybe_written = OrderedSet(self.maybe_written)
 
-        for child in ast.get_child_nodes(node):
-            ast.walk(child, self)
+        for child in ast.iter_child_nodes(node):
+            self.visit(child)
 
         comp_names = list(
             chain.from_iterable(
@@ -890,7 +939,7 @@ class _FunctionInformationCollector(object):
         elif isinstance(node, ast.Name):
             yield node.id
         else:
-            assert False, "Unexpected node type in list comprehension target: %s" % node
+            assert False, f"Unexpected node type in list comprehension target: {node!r}"
 
     def _If(self, node):
         self._handle_conditional_node(node)
@@ -902,18 +951,18 @@ class _FunctionInformationCollector(object):
     def _For(self, node):
         with self._handle_loop_context(node), self._handle_conditional_context(node):
             # iter has to be checked before the target variables
-            ast.walk(node.iter, self)
-            ast.walk(node.target, self)
+            self.visit(node.iter)
+            self.visit(node.target)
 
             for child in node.body:
-                ast.walk(child, self)
+                self.visit(child)
             for child in node.orelse:
-                ast.walk(child, self)
+                self.visit(child)
 
     def _handle_conditional_node(self, node):
         with self._handle_conditional_context(node):
-            for child in ast.get_child_nodes(node):
-                ast.walk(child, self)
+            for child in ast.iter_child_nodes(node):
+                self.visit(child)
 
     @contextmanager
     def _handle_conditional_context(self, node):
@@ -935,19 +984,18 @@ class _FunctionInformationCollector(object):
 
 
 def _get_argnames(arguments):
-    result = [
-        pycompat.get_ast_arg_arg(node)
-        for node in arguments.args
-        if isinstance(node, pycompat.ast_arg_type)
-    ]
+    result = []
+    result.extend(node.arg for node in getattr(arguments, "posonlyargs", []))
+    result.extend(node.arg for node in arguments.args)
     if arguments.vararg:
-        result.append(pycompat.get_ast_arg_arg(arguments.vararg))
+        result.append(arguments.vararg.arg)
     if arguments.kwarg:
-        result.append(pycompat.get_ast_arg_arg(arguments.kwarg))
+        result.append(arguments.kwarg.arg)
+    result.extend(node.arg for node in arguments.kwonlyargs)
     return result
 
 
-class _VariableReadsAndWritesFinder(object):
+class _VariableReadsAndWritesFinder(ast.RopeNodeVisitor):
     def __init__(self):
         self.written = set()
         self.read = set()
@@ -961,8 +1009,8 @@ class _VariableReadsAndWritesFinder(object):
     def _FunctionDef(self, node):
         self.written.add(node.name)
         visitor = _VariableReadsAndWritesFinder()
-        for child in ast.get_child_nodes(node):
-            ast.walk(child, visitor)
+        for child in ast.iter_child_nodes(node):
+            visitor.visit(child)
         self.read.update(visitor.read - visitor.written)
 
     def _Class(self, node):
@@ -974,7 +1022,7 @@ class _VariableReadsAndWritesFinder(object):
             return set(), set()
         node = _parse_text(code)
         visitor = _VariableReadsAndWritesFinder()
-        ast.walk(node, visitor)
+        visitor.visit(node)
         return visitor.read, visitor.written
 
     @staticmethod
@@ -983,18 +1031,18 @@ class _VariableReadsAndWritesFinder(object):
             return set(), set()
         node = _parse_text(code)
         visitor = _VariableReadsAndWritesFinder()
-        ast.walk(node, visitor)
+        visitor.visit(node)
         return visitor.read
 
 
-class _BaseErrorFinder(object):
+class _BaseErrorFinder(ast.RopeNodeVisitor):
     @classmethod
     def has_errors(cls, code):
         if code.strip() == "":
             return False
         node = _parse_text(code)
         visitor = cls()
-        ast.walk(node, visitor)
+        visitor.visit(node)
         return visitor.error
 
 
@@ -1012,14 +1060,14 @@ class _UnmatchedBreakOrContinueFinder(_BaseErrorFinder):
     def loop_encountered(self, node):
         self.loop_count += 1
         for child in node.body:
-            ast.walk(child, self)
+            self.visit(child)
         self.loop_count -= 1
         if node.orelse:
             if isinstance(node.orelse, (list, tuple)):
                 for node_ in node.orelse:
-                    ast.walk(node_, self)
+                    self.visit(node_)
             else:
-                ast.walk(node.orelse, self)
+                self.visit(node.orelse)
 
     def _Break(self, node):
         self.check_loop()
@@ -1055,7 +1103,35 @@ class _AsyncStatementFinder(_BaseErrorFinder):
         pass
 
 
-class _GlobalFinder(object):
+class _UnbalancedRegionFinder(_BaseErrorFinder):
+    """
+    Flag an error if we are including the start of a block without the end.
+    We detect this by ensuring there is no AST node that starts inside the
+    selected range but ends outside of it.
+    """
+
+    def __init__(self, line_start: int, line_end: int):
+        self.error = False
+        self.line_start = line_start
+        self.line_end = line_end
+
+    def generic_visit(self, node: ast.AST):
+        if not hasattr(node, "end_lineno"):
+            super().generic_visit(node)  # Visit children
+            return
+        ends_before_range_starts = node.end_lineno < self.line_start
+        starts_after_range_ends = node.lineno > self.line_end
+        if ends_before_range_starts or starts_after_range_ends:
+            return  # Don't visit children
+        starts_on_or_after_range_start = node.lineno >= self.line_start
+        ends_after_range_ends = node.end_lineno > self.line_end
+        if starts_on_or_after_range_start and ends_after_range_ends:
+            self.error = True
+            return  # Don't visit children
+        super().generic_visit(node)  # Visit children
+
+
+class _GlobalFinder(ast.RopeNodeVisitor):
     def __init__(self):
         self.globals_ = OrderedSet()
 
@@ -1063,6 +1139,14 @@ class _GlobalFinder(object):
         self.globals_.add(*node.names)
 
 
+class _NonlocalFinder(ast.RopeNodeVisitor):
+    def __init__(self):
+        self.nonlocals_ = OrderedSet()
+
+    def _Nonlocal(self, node):
+        self.nonlocals_.add(*node.names)
+
+
 def _get_function_kind(scope):
     return scope.pyobject.get_kind()
 
@@ -1103,6 +1187,10 @@ def _get_single_expression_body(extracted, info):
     if not large_multiline:
         extracted = _join_lines(extracted)
     multiline_expression = "\n" in extracted
-    if info.returning_named_expr or (multiline_expression and not large_multiline):
+    if (
+        info.returning_named_expr
+        or info.returning_generator_exp
+        or (multiline_expression and not large_multiline)
+    ):
         extracted = "(" + extracted + ")"
     return extracted
